import numpy
from matplotlib import pyplot

from netgen import csg
import ngsolve

from netgen import gui

# Constants
π = numpy.pi
c0 = 299792458
μ0 = 4e-7 * π
ε0 = 1.0 / (c0 * c0 * μ0)
η0 = c0 * μ0

# Parameters
λ = 1.55
k0 = 2 * π / λ
n1 = 1
n2 = 1.5

domain1 = csg.OrthoBrick(csg.Pnt(-λ, -λ, -λ), csg.Pnt(0, λ, λ))
domain2 = csg.OrthoBrick(csg.Pnt(0, -λ, -λ), csg.Pnt(λ, λ, λ))
geo = csg.CSGeometry()
geo.Add(domain1.mat("dom1"))
geo.Add(domain2.mat("dom2"), maxh=λ / n2 / 5)
mesh = ngsolve.Mesh(geo.GenerateMesh(maxh=λ / n1 / 5))

geo.Draw()

# Boundary-value problem
fes = ngsolve.HCurl(mesh, order=2, complex=True)
print("DOF:", fes.ndof)

solve_for = "E"

if solve_for == "E":
    # Electric formulation
    ξ = η0
    p = 1
    q = ngsolve.CoefficientFunction(
        [n1 ** 2 if m == "dom1" else n2 ** 2 for m in mesh.GetMaterials()]
    )
    q.Compile(True)
    ngsolve.Draw(q, mesh, "q")
else:
    # Magnetic formulation
    ξ = 1 / η0
    q = 1
    p = ngsolve.CoefficientFunction(
        [1 / n1 ** 2 if m == "dom1" else 1 / n2 ** 2 for m in mesh.GetMaterials()]
    )
    p.Compile(True)
    ngsolve.Draw(p, mesh, "p")

# Source
u0 = None
fu = ngsolve.CoefficientFunction(
    (
        0,
        ngsolve.exp(
            -(ngsolve.x ** 2 + ngsolve.y ** 2 + ngsolve.z ** 2) / (2 * (λ / 4) ** 2)
        ),
        0,
    )
)
fu.Compile(True)
ngsolve.Draw(fu, mesh, "Fu")

# Problem setup
u, w = fes.TnT()

# Bilinear form
a = ngsolve.BilinearForm(fes)
a += p * (ngsolve.curl(u) * ngsolve.curl(w)) * ngsolve.dx
a += -(k0 ** 2) * q * (u * w) * ngsolve.dx

apre = ngsolve.BilinearForm(fes)
apre += p * (ngsolve.curl(u) * ngsolve.curl(w)) * ngsolve.dx
apre += -(1+1j)*(k0 ** 2) * q * (u * w) * ngsolve.dx


# Linear form
b = ngsolve.LinearForm(fes)
b += -1j * k0 * ξ * (fu * w) * ngsolve.dx

# Preconditioner
pre = ngsolve.Preconditioner(apre, "bddc")

# Solution
sol_u = ngsolve.GridFunction(fes, name=solve_for)

with ngsolve.TaskManager():
    print("Assembling a…", flush=True)
    a.Assemble()
    apre.Assemble()
    print("Assembling b…", flush=True)
    b.Assemble()
    print("Solving…", flush=True)
    solver = ngsolve.krylovspace.CGSolver(
        a.mat,
        pre.mat,
        conjugate=False,
        maxsteps=100000,
        tol=0,
        abstol=1e-12,
        printing=True,
    )
    solver.Solve(b.vec, sol_u.vec, False)
    # sol_u.vec.data = a.mat.Inverse(inverse="sparsecholesky") * b.vec

# print(f"Error: {solver.errors[-1]:.3g}  iterations: {len(solver.errors)}")

# ngsolve.Draw(sol_u)
for i in range(3):
    ngsolve.Draw(sol_u[i], mesh, solve_for + "xyz"[i])
